﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using ZOIL.DataModel;
using System.Windows.Interactivity;
using System.Windows;

namespace ZOIL.Library.Behaviors
{
    public class ConnectionBehavior : Behavior<ZLandscapeCanvas>
    {
        private Dictionary<ILink, ShapeToShapeConnection> _establishedLinks;

        private IDisposable _elementAddedSubscription;
        private IDisposable _elementRemovedSubscription;

        public ConnectionBehavior()
        {
            _establishedLinks = new Dictionary<ILink, ShapeToShapeConnection>();
        }


        #region AreConnectionsEnabled

        /// <summary>
        /// AreConnectionsEnabled Dependency Property
        /// </summary>
        public static readonly DependencyProperty AreConnectionsEnabledProperty =
            DependencyProperty.Register("AreConnectionsEnabled", typeof(bool), typeof(ConnectionBehavior),
                new FrameworkPropertyMetadata((bool)true,
                    new PropertyChangedCallback(OnAreConnectionsEnabledChanged)));

        /// <summary>
        /// Gets or sets the AreConnectionsEnabled property. This dependency property 
        /// indicates ....
        /// </summary>
        public bool AreConnectionsEnabled
        {
            get { return (bool)GetValue(AreConnectionsEnabledProperty); }
            set { SetValue(AreConnectionsEnabledProperty, value); }
        }

        /// <summary>
        /// Handles changes to the AreConnectionsEnabled property.
        /// </summary>
        private static void OnAreConnectionsEnabledChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            ConnectionBehavior target = (ConnectionBehavior)d;
            bool oldAreConnectionsEnabled = (bool)e.OldValue;
            bool newAreConnectionsEnabled = target.AreConnectionsEnabled;
            target.OnAreConnectionsEnabledChanged(oldAreConnectionsEnabled, newAreConnectionsEnabled);
        }

        /// <summary>
        /// Provides derived classes an opportunity to handle changes to the AreConnectionsEnabled property.
        /// </summary>
        protected virtual void OnAreConnectionsEnabledChanged(bool oldAreConnectionsEnabled, bool newAreConnectionsEnabled)
        {
        }

        #endregion

        protected override void OnAttached()
        {
            var onElementAdded = Observable.FromEvent<ZLandscapeCanvasEventArgs>(AssociatedObject, "ElementAdded").Where(evt => evt.EventArgs.Model is ILinkable).Select(evt => new Tuple<ILinkable, FrameworkElement>(evt.EventArgs.Model as ILinkable, evt.EventArgs.View)).Where(tuple => tuple.Item1.Links.Count() > 0);
            _elementAddedSubscription = onElementAdded.Subscribe(CreateLinks);

            var onElementRemoved = Observable.FromEvent<ZLandscapeCanvasEventArgs>(AssociatedObject, "ElementRemoved").Do(evt => Console.WriteLine("Element Removed")).Where(evt => evt.EventArgs.Model is ILinkable).Select(evt => new Tuple<ILinkable, FrameworkElement>(evt.EventArgs.Model as ILinkable, evt.EventArgs.View)).Where(tuple => tuple.Item1.Links.Count() > 0);
            _elementRemovedSubscription = onElementRemoved.Subscribe(RemoveLinks);
        }

        private void RemoveLinks(Tuple<ILinkable, FrameworkElement> tuple)
        {
            ILinkable node = tuple.Item1;

            foreach (ILink link in node.Links)
            {
                if (_establishedLinks.ContainsKey(link))
                {
                    AssociatedObject.RemoveNonPersistableItem(_establishedLinks[link]);
                    _establishedLinks.Remove(link);
                }
            }
        }

        private void CreateLinks(Tuple<ILinkable, FrameworkElement> tuple)
        {
            ILinkable node = tuple.Item1;

            foreach(ILink link in node.Links)
            {
                if(!_establishedLinks.ContainsKey(link))
                {
                    IPersistable node1 = link.Node1;
                    IPersistable node2 = link.Node2;

                    IConnectionShape shape1 = MVVMUtil.getViewModel(node1) as IConnectionShape;
                    IConnectionShape shape2 = MVVMUtil.getViewModel(node2) as IConnectionShape;

                    if(shape1 != null && shape2 != null)
                    {
                        ShapeToShapeConnection connection = new ShapeToShapeConnection(shape1, shape2);   
                        _establishedLinks.Add(link, connection);
                        AssociatedObject.AddNonPersistableItem(connection, VisualProperties.Default());
                    }
                }
            }
        }

    }
}
