﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Surface.Presentation;
using Microsoft.Surface.Presentation.Manipulations;
using ZOIL.Framework.Presentation;
using System.Windows.Controls;
using System.Windows;
using System.Windows.Shapes;
using System.Windows.Media;
using ZOIL.Framework.Core;

namespace ZOIL.Library.Behaviors
{
    public static class ZObjectSurfaceHandler
    {
        private static Dictionary<ZComponent, ZObjectManipulationInformation> ElementsBeingManipulated = new Dictionary<ZComponent, ZObjectManipulationInformation>();
        private static Dictionary<Affine2DInertiaProcessor, ZObjectManipulationInformation> ElementsBeingAccelerated = new Dictionary<Affine2DInertiaProcessor, ZObjectManipulationInformation>();

        private static Dictionary<FrameworkElement, Contact> _contactsDown = new Dictionary<FrameworkElement, Contact>();

        public static void OnContactTapGesture(object sender, ContactEventArgs e)
        {
            if (!e.Contact.IsFingerRecognized)
            {
                return;
            }

            //possible Multiple Click Zoom
            if (_contactsDown.Count > 1)
            {
                return;
            }

            FrameworkElement zoomTarget = ZTreeHelper.GetClosestZoomTarget(sender as FrameworkElement, e.Source as FrameworkElement);
            if (zoomTarget != null)
            {
                ZInformationLandscape landscape = ZTreeHelper.TryFindParent<ZInformationLandscape>(sender as DependencyObject);
                if (landscape != null)
                {
                    landscape.ZoomOnto(zoomTarget, false, true);
                    zoomTarget.RaiseEvent(new RoutedEventArgs(ZInformationLandscape.ClickZoomEvent, zoomTarget));
                }
            }
            e.Handled = true;
        }

        public static void OnContactDown(object sender, ContactEventArgs e)
        {
            if (!e.Contact.IsFingerRecognized)
                return;

            ZComponent comp = sender as ZComponent;

            if (comp != null)
            {
                Panel.SetZIndex(comp, ZComponent.ZIndexCounter++);

                if (comp.InertiaProcessor != null && comp.InertiaProcessor.IsRunning)
                {
                    try
                    {
                        comp.InertiaProcessor.End();
                    }
                    catch
                    {
                        //do nothing
                    }
                }

                if (!ElementsBeingManipulated.ContainsKey(comp))
                {
                    ZObjectManipulationInformation info = new ZObjectManipulationInformation(comp);
                    info.ElementBeingDragged = ZBehavior.FindCanvasChild(sender as UIElement);
                    info.OriginalPositionOnWindow = e.GetPosition(null);
                    
                    info.OriginalFrameworkElement = ((e.Source as FrameworkElement != null) ? e.Source as FrameworkElement : info.ElementBeingDragged as FrameworkElement);

                    // TODO: replace using TransformToAncestor()
                    info.ParentLandscapes = ZTreeHelper.TryFindParents<ZInformationLandscape>(comp);
                    info.Left = Canvas.GetLeft(comp);
                    info.Top = Canvas.GetTop(comp);
                    info.Contacts.Add(e.Contact);

                    info.OriginalPositionOnCanvas = new Point(info.Left, info.Top);

                    ElementsBeingManipulated[comp] = info;
                    comp.ManipulationProcessor = null;
                    InitProcessor(comp, ZObjectDragDropBehavior.GetIsDraggable(comp), ZObjectResizeBehavior.GetIsResizable(comp), ZObjectRotateBehavior.GetIsRotatable(comp));

                    comp.ManipulationProcessor.Affine2DManipulationDelta += new EventHandler<Affine2DOperationDeltaEventArgs>(ManipulationProcessor_Affine2DManipulationDelta);
                    comp.ManipulationProcessor.Affine2DManipulationCompleted += new EventHandler<Affine2DOperationCompletedEventArgs>(ManipulationProcessor_Affine2DManipulationCompleted);
                }
                else
                {
                    ZObjectManipulationInformation info = ElementsBeingManipulated[comp];
                    info.Contacts.Add(e.Contact);
                }
                comp.ManipulationProcessor.BeginTrack(e.Contact);
                Contacts.CaptureContact(e.Contact, comp);
            }
            e.Handled = true;
        }

        public static void OnContactHold(object sender, ContactEventArgs e)
        {
            if (!e.Contact.IsFingerRecognized)
                return;

            FrameworkElement fe = sender as FrameworkElement;
            //if fe is a framework element, is not in the list yet and is child of a landscape, then add...
            if (fe != null && !_contactsDown.Keys.Contains(fe) && ZTreeHelper.TryFindParent<ZInformationLandscape>(fe) != null)
            {
                Canvas.SetZIndex(fe, ZComponent.ZIndexCounter++);
                _contactsDown.Add(fe, e.Contact);
            }
        }

        public static void OnContactUp(object sender, ContactEventArgs e)
        {
            if (_contactsDown.Count >= 2 && _contactsDown.Values.Contains(e.Contact))
            {
                List<FrameworkElement> list = new List<FrameworkElement>();
                foreach (FrameworkElement fe in _contactsDown.Keys)
                {
                    list.Add(fe);
                }

                
                _contactsDown.Clear();

            }

            if (_contactsDown.Keys.Contains(sender as FrameworkElement) && (e.Contact == _contactsDown[sender as FrameworkElement]))
            {
                _contactsDown.Remove(sender as FrameworkElement);
            }

        }

        private static void InitProcessor(ZComponent comp, bool supportsDragging, bool supportsResizing, bool supportsRotating)
        {
            Affine2DManipulations manipulations = Affine2DManipulations.None;

            if (supportsDragging)
            {
                manipulations |= Affine2DManipulations.TranslateX;
                manipulations |= Affine2DManipulations.TranslateY;
            }
            if (supportsResizing)
            {
                manipulations |= Affine2DManipulations.Scale;
            }
            if (supportsRotating)
            {
                manipulations |= Affine2DManipulations.Rotate;
            }

            if (comp.ManipulationProcessor == null)
            {
                comp.ManipulationProcessor = new Affine2DManipulationProcessor(manipulations, comp, true);
            }
            else
            {
                comp.ManipulationProcessor.SupportedManipulations = manipulations;
            }
        }



        #region manipulation

        static void ManipulationProcessor_Affine2DManipulationDelta(object sender, Affine2DOperationDeltaEventArgs e)
        {
            Affine2DManipulationProcessor processor = sender as Affine2DManipulationProcessor;

            ZComponent comp = processor.Container as ZComponent;
            if (comp != null)
            {
                if (ElementsBeingManipulated.ContainsKey(comp))
                {
                    ZObjectManipulationInformation info = ElementsBeingManipulated[comp];

                    if (comp != info.ComponentBeingManipulated)
                    {
#if DEBUG
                        ZDebug.Log(DebugLevel.Warning, DebugCategories.Components, "Warning, treating a component in two different ways!");
#endif
                    }


                    // TODO: dragging works well with following code, without own dragging method --> 
                    // evaluate flag for dragging true/false somewhere below
                    if (ZObjectResizeBehavior.GetIsResizable(comp) || ZObjectDragDropBehavior.GetIsDraggable(comp))
                    {
                        #region Fire Drag Events, TODO: extract this to extra method
                        if (ZObjectDragDropBehavior.GetIsDraggable(comp) && e.ScaleDelta == 1 && e.CumulativeTranslation.Length > 10)
                        {
                            //if we are not dragging yet fire drag started event and set is draggiing to true
                            if (!ZObjectDragDropBehavior.GetIsDragging(comp))
                            {
                                ZObjectDragDropEventArgs args = new ZObjectDragDropEventArgs();
                                args.RoutedEvent = ZObjectDragDropBehavior.DragStartedEvent;
                                args.Source = comp;
                                args.CurrentDragPoint = e.ManipulationOrigin;
                                comp.RaiseEvent(args);
                                ZObjectDragDropBehavior.SetIsDragging(comp, true);
                            }
                            //otherwise fire only drag delta event
                            else
                            {
                                if (info.DeltaFired)
                                {
                                    info.DeltaFired = false;
                                }
                                else
                                {
                                    info.DeltaFired = true;
                                    //Raise DragDeltaEvent
                                    ZObjectDragDropEventArgs args = new ZObjectDragDropEventArgs();
                                    args.CurrentDragPoint = e.ManipulationOrigin;
                                    args.RoutedEvent = ZObjectDragDropBehavior.DragDeltaEvent;
                                    args.Source = comp;
                                    comp.RaiseEvent(args);
                                }
                            }
                        }
                        else
                        {
                            ZObjectDragDropBehavior.SetIsDragging(comp, false);
                        }
                        #endregion

                        Point Center = new Point(Canvas.GetLeft(comp) + comp.Width / 2, Canvas.GetTop(comp) + comp.Height / 2);

                        #region Visual Debug
                        // YELLOW: Center -> visual point for debugging
                        //Ellipse cen = new Ellipse();
                        //cen.Width = 10;
                        //cen.Height = 10;
                        //cen.Fill = new SolidColorBrush(Colors.Yellow);
                        //((Canvas)comp.ParentInformationLandscape.Content).Children.Add(cen);
                        //Canvas.SetLeft(cen, Center.X);
                        //Canvas.SetTop(cen, Center.Y);
                        #endregion

                        // TODO: add nested landscape support
                        double ILScale = info.getScaleFactor();
                        Vector deltaXY = e.Delta / ILScale;

                        // gets manipulation origin in screen(!) coordinates
                        Point lastManipulationOrigin = e.ManipulationOrigin;
                        // convert screen coordinates to window coordinates
                        lastManipulationOrigin = Window.GetWindow(comp).PointFromScreen(lastManipulationOrigin);
                        // convert window coordinates to landscape coordinates
                        lastManipulationOrigin = comp.ParentInformationLandscape.TransformWindowToLandscapeCoordinates(lastManipulationOrigin);
                        lastManipulationOrigin -= deltaXY;

                        #region VisualDebug
                        // RED: LastManipulationOrigin -> visual point for debugging
                        //Ellipse lmo = new Ellipse();
                        //lmo.Width = 10;
                        //lmo.Height = 10;
                        //lmo.Fill = new SolidColorBrush(Colors.Red);                          
                        //((Canvas)comp.ParentInformationLandscape.Content).Children.Add(lmo);
                        //Canvas.SetLeft(lmo, lastManipulationOrigin.X);
                        //Canvas.SetTop(lmo, lastManipulationOrigin.Y);
                        #endregion

                        Vector vector = (Vector)(Center - lastManipulationOrigin);
                        
                        double scale = 1.0;
                        if (ZObjectResizeBehavior.GetIsResizable(comp))
                        {
                            scale = e.ScaleDelta;
                        }

                        Vector vector4 = (Vector)(vector * scale);
                        deltaXY += vector4 - vector;

                        //components size is adjusted to new scale
                        comp.Width *= scale;
                        comp.Height *= scale;

                        //look for a translation during resizing and adjust the component position
                        if (deltaXY.LengthSquared > 0.0)
                        {
                            Point center = Center;
                            center.X += deltaXY.X;
                            center.Y += deltaXY.Y;
                            Center = center;

                            Canvas.SetLeft(comp, Center.X - comp.Width / 2.0);
                            Canvas.SetTop(comp, Center.Y - comp.Height / 2.0);
                        }

                        #region Visual Debug
                        // GREEN: new center -> visual point for debugging
                        //Ellipse newcen = new Ellipse();
                        //newcen.Width = 10;
                        //newcen.Height = 10;
                        //newcen.Fill = new SolidColorBrush(Colors.Green);
                        //((Canvas)comp.ParentInformationLandscape.Content).Children.Add(newcen);
                        //Canvas.SetLeft(newcen, Center.X);
                        //Canvas.SetTop(newcen, Center.Y);
                        #endregion
                    }

                    //if component is rotated adjust the angle value
                    if (ZObjectRotateBehavior.GetIsRotatable(comp))
                    {
                        comp.Angle += e.RotationDelta;
                    }

                    /* //TODO: for flexibility, separate drag and resize, currently this code is inside resizing
                    if (ZObjectDragDropBehavior.GetIsDraggable(comp) && e.ScaleDelta == 1)
                    {
                        //if we are not dragging yet fire drag started event and set is draggiing to true
                        if (!ZObjectDragDropBehavior.GetIsDragging(comp))
                        {
                            ZObjectDragDropEventArgs args = new ZObjectDragDropEventArgs();
                            args.RoutedEvent = ZObjectDragDropBehavior.DragStartedEvent;
                            args.Source = comp;
                            args.CurrentDragPoint = e.ManipulationOrigin;
                            comp.RaiseEvent(args);
                            ZObjectDragDropBehavior.SetIsDragging(comp, true);
                        }
                        //otherwise fire only drag delta event
                        else
                        {
                            //Raise DragDeltaEvent
                            ZObjectDragDropEventArgs args = new ZObjectDragDropEventArgs();
                            args.CurrentDragPoint = e.ManipulationOrigin;
                            args.RoutedEvent = ZObjectDragDropBehavior.DragDeltaEvent;
                            args.Source = comp;
                            comp.RaiseEvent(args);
                        }

                        double scale = info.getScaleFactor();

                        Canvas.SetLeft(comp, info.Left + e.CumulativeTranslation.X / scale);
                        Canvas.SetTop(comp, info.Top + e.CumulativeTranslation.Y / scale);
                    }

                 */
                }
            }
        }

        static void ManipulationProcessor_Affine2DManipulationCompleted(object sender, Affine2DOperationCompletedEventArgs e)
        {
            Affine2DManipulationProcessor processor = sender as Affine2DManipulationProcessor;

            ZComponent comp = processor.Container as ZComponent;
            if (comp != null)
            {
                if (ElementsBeingManipulated.ContainsKey(comp))
                {
                    ZObjectManipulationInformation info = ElementsBeingManipulated[comp];
                    if (comp.PhysicsEnabled)
                    {
                        Point currentPositionOnCanvas = new Point(Canvas.GetLeft(comp), Canvas.GetTop(comp));
                        double distance = Point.Subtract(info.OriginalPositionOnCanvas, currentPositionOnCanvas).Length;

                        if (distance < 20)
                        {
                            ElementsBeingManipulated.Remove(comp);
                            return;
                        }

                        ElementsBeingAccelerated[comp.InertiaProcessor] = info;
                        ElementsBeingManipulated.Remove(comp);

                        try
                        {
                            info.Left = Canvas.GetLeft(comp);
                            info.Top = Canvas.GetTop(comp);
                            comp.InertiaProcessor.InitialOrigin = e.ManipulationOrigin;
                            comp.InertiaProcessor.InitialVelocity = e.Velocity;
                            comp.InertiaProcessor.InitialAngularVelocity = e.AngularVelocity;
                            comp.InertiaProcessor.Affine2DInertiaDelta += new EventHandler<Affine2DOperationDeltaEventArgs>(InertiaProcessor_Affine2DInertiaDelta);
                            comp.InertiaProcessor.Affine2DInertiaCompleted += new EventHandler<Affine2DOperationCompletedEventArgs>(InertiaProcessor_Affine2DInertiaCompleted);
                            comp.InertiaProcessor.DesiredAngularDeceleration = 0.00005;
                            comp.InertiaProcessor.DesiredDeceleration = 0.001;
                            comp.InertiaProcessor.ElasticMargin = new System.Windows.Thickness(50);
                            comp.InertiaProcessor.Bounds = new System.Windows.Thickness(200, 200, Window.GetWindow(comp).ActualWidth - 200, Window.GetWindow(comp).ActualHeight - 200);
                        }
                        catch
                        {
                            if (comp != null)
                            {
                                Console.WriteLine("Unknown object is null, see Inertia!");
                            }
                        }
                        try
                        {
                            comp.InertiaProcessor.Begin();
                        }
                        catch
                        {
                            ZObjectDragDropEventArgs args = new ZObjectDragDropEventArgs();
                            args.RoutedEvent = ZObjectDragDropBehavior.DragFinishedEvent;
                            args.Source = info.ComponentBeingManipulated;
                            args.CurrentDragPoint = e.ManipulationOrigin;
                            info.ComponentBeingManipulated.RaiseEvent(args);
                        }
                    }
                }
            }
        }

        #endregion

        #region intertia

        static void InertiaProcessor_Affine2DInertiaCompleted(object sender, Affine2DOperationCompletedEventArgs e)
        {
            Affine2DInertiaProcessor processor = sender as Affine2DInertiaProcessor;

            if (ElementsBeingAccelerated.ContainsKey(processor))
            {
                ZObjectManipulationInformation info = ElementsBeingAccelerated[processor];
                //Raise DragFinishedEvent and set IsDragging to false
                ZObjectDragDropEventArgs args = new ZObjectDragDropEventArgs();
                args.RoutedEvent = ZObjectDragDropBehavior.DragFinishedEvent;
                args.Source = info.ComponentBeingManipulated;
                args.CurrentDragPoint = e.ManipulationOrigin;
                info.ComponentBeingManipulated.RaiseEvent(args);
                //ZObjectDragDropBehavior.SetIsDragging(info.ComponentBeingManipulated, false);

                ElementsBeingAccelerated.Remove(processor);
            }
        }

        static void InertiaProcessor_Affine2DInertiaDelta(object sender, Affine2DOperationDeltaEventArgs e)
        {
            Affine2DInertiaProcessor processor = sender as Affine2DInertiaProcessor;

            if (ElementsBeingAccelerated.ContainsKey(processor))
            {
                ZObjectManipulationInformation info = ElementsBeingAccelerated[processor];

                ZComponent comp = info.ComponentBeingManipulated;

                if (ZObjectDragDropBehavior.GetIsDraggable(comp))
                {
                    //Raise DragDeltaEvent
                    ZObjectDragDropEventArgs args = new ZObjectDragDropEventArgs();
                    args.CurrentDragPoint = e.ManipulationOrigin;
                    args.RoutedEvent = ZObjectDragDropBehavior.DragDeltaEvent;
                    args.Source = comp;
                    comp.RaiseEvent(args);

                    double scale = info.getScaleFactor();

                    Canvas.SetLeft(comp, info.Left + e.CumulativeTranslation.X / scale);
                    Canvas.SetTop(comp, info.Top + e.CumulativeTranslation.Y / scale);
                }

                if (ZObjectRotateBehavior.GetIsRotatable(comp))
                {
                    comp.Angle += e.RotationDelta;
                }
            }
        }


        #endregion
    }
}
