﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using ZOIL.DataBackend;
using ZOIL.DataModel;
using ZOIL.Framework.Presentation;
using Db4objects.Db4o.Linq;
using ZOIL.Library.Behaviors;
using ZOIL.Framework.Core;

namespace ZOIL.Library.Components
{
    [ViewModelFor(typeof(ZPortalShapeView))]
    public class ZPortalShapeViewModel : BaseViewModel, IConnectionShape
    {
        #region ctor

        public ZPortalShapeViewModel()
            : base()
        {
            RegisterChangeNotification("ShapeGeometry", "ShapeGeometry");
            initCommandBindings();
            ConnectionShape = new RectangleGeometry();

            _shapeMoves = new Subject<Unit>();
            _shapeMoveSubscription = _shapeMoves.Sample(TimeSpan.FromMilliseconds(50)).ObserveOnDispatcher().Subscribe(unit => OnShapeMoved());
        }

        #endregion

        #region private fields

        private ZInformationLandscape _landscape;
        private IDisposable _shapeMoveSubscription;
        private Subject<Unit> _shapeMoves;

        #endregion

        #region public properties

        #region SelectedElements

        public DObservableCollection<Object> SelectedElements
        {
            get
            {
                return (Model as ZPortalShape).SelectedElements;
            }
            set
            {
                (Model as ZPortalShape).SelectedElements = value;
                RaisePropertyChangedEvent("SelectedElements");
            }
        }

        #endregion

        #region ShapeGeometry

        private Geometry _shapeGeometry;

        public Geometry ShapeGeometry
        {
            get
            {
                if(_shapeGeometry == null)
                    _shapeGeometry = (Model as ZPortalShape).ShapeGeometry;
                return _shapeGeometry;
            }
            set
            {
                _shapeGeometry = value;
                (Model as ZPortalShape).ShapeGeometry = value;
            }
        }

        #endregion

        #region HitTestType

        public Type HitTestType
        {
            get
            {
                return (Model as ZPortalShape).HitTestType;
            }
            set
            {
                (Model as ZPortalShape).HitTestType = value;
                RaisePropertyChangedEvent("HitTestType");
            }
        }

        #endregion

        #region ConnectionShape

        private Geometry _connectionShape;

        /// <summary>
        /// Gets or sets the ConnectionShape property. This observable property 
        /// indicates
        /// </summary>
        public Geometry ConnectionShape
        {
            get { return _connectionShape; }
            set
            {
                if (_connectionShape != value)
                {
                    Geometry old = _connectionShape;
                    _connectionShape = value;
                    OnConnectionShapeChanged(old, value);
                    RaisePropertyChangedEvent("ConnectionShape");
                }
            }
        }

        /// <summary>
        /// Provides derived classes an opportunity to handle changes to the ConnectionShape property.
        /// </summary>
        protected virtual void OnConnectionShapeChanged(Geometry oldConnectionShape, Geometry newConnectionShape)
        {
        }

        #endregion

        #endregion

        #region private methods

        private void InitShape()
        {
            _landscape = ZTreeHelper.TryFindParent<ZInformationLandscape>(View);
            ZPortalShape shape = (Model as ZPortalShape);
            
            if (!shape.IsInitialized)
            {
                var rootCollections = from RootCollection rc in Database.getInstance().Objects
                                      select rc;

                RootCollection _rootCollection = rootCollections.Single();
                IVisualProperties shapeProp = _rootCollection.GetVisualProperties(shape.ID).First();
                shapeProp.Width = shape.Bounds.Width;
                shapeProp.Height = shape.Bounds.Height;
                shapeProp.X = shape.Bounds.Left;
                shapeProp.Y = shape.Bounds.Top;
                //shapeProp.TemplateId = "PortalShapeTemplate";
                shape.IsInitialized = true;
            }
            
            OnShapeMoved();
        }

        private void initCommandBindings()
        {
            this.MoveShapeCommand = new RoutedCommand("MoveShapeCommand", typeof(ZPortalShapeViewModel));
            CommandBinding cb = new CommandBinding(MoveShapeCommand, ExecuteMoveShapeCommand, CanExecuteMoveShapeCommand);
            this.CommandBindings.Add(cb);

            this.FinishedMoveShapeCommand = new RoutedCommand("FinishedMoveShapeCommand", typeof(ZPortalShapeViewModel));
            CommandBinding cb2 = new CommandBinding(FinishedMoveShapeCommand, ExecuteFinishedMoveShapeCommand, CanExecuteFinishedMoveShapeCommand);
            this.CommandBindings.Add(cb2);

            this.ElementAddedCommand = new RoutedCommand("ElementAddedCommand", typeof(ZPortalViewModel));
            CommandBinding cb3 = new CommandBinding(ElementAddedCommand, ExecuteElementAddedCommand, CanExecuteElementAddedCommand);
            this.CommandBindings.Add(cb3);
        }

        private void RecalculateShapeGeometry()
        {
            double offsetx = Canvas.GetLeft(View) - (Model as ZPortalShape).Bounds.Left;
            double offsety = Canvas.GetTop(View) - (Model as ZPortalShape).Bounds.Top;

            ShapeGeometry.Transform = new TranslateTransform(offsetx, offsety);
            ShapeGeometry = ShapeGeometry.GetOutlinedPathGeometry();
        }

        private void DoHitTest()
        {
            if (_landscape == null)
                _landscape = ZTreeHelper.TryFindParent<ZInformationLandscape>(View);

            TransformGroup landscapeTransform = _landscape.LandscapeTransform;

            Geometry HitTestGeometry = ShapeGeometry.Clone();
            HitTestGeometry.Transform = landscapeTransform;

            var hitTestResult = ZHitTestHelper.GetElementsInGeometry(HitTestGeometry, _landscape, HitTestType).OfType<ZComponent>().Select(element =>
            {
                IViewModel vm = element.DataContext as IViewModel;
                return vm.Model;
            });

            IEnumerable<Object> elementsToBeRemoved = SelectedElements.Except(hitTestResult).ToArray();
            IEnumerable<Object> elementsToBeAdded = hitTestResult.Except(SelectedElements).ToArray();

            SelectedElements.RemoveRange(elementsToBeRemoved);
            SelectedElements.AddRange(elementsToBeAdded);
        }

        #endregion

        #region event handling

        private void OnShapeMoved()
        {
            RecalculateShapeGeometry();
            DoHitTest();
            ConnectionShape = ShapeGeometry;
        }

        #endregion

        #region commands

        #region Finished Move Shape Command

        private RoutedCommand _finishedMoveShapeCommand;
        public RoutedCommand FinishedMoveShapeCommand
        {
            get
            {
                return _finishedMoveShapeCommand;
            }
            set
            {
                _finishedMoveShapeCommand = value;
                RaisePropertyChangedEvent("FinishedMoveShapeCommand");
            }
        }

        private void CanExecuteFinishedMoveShapeCommand(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = true;
        }

        private void ExecuteFinishedMoveShapeCommand(object sender, ExecutedRoutedEventArgs e)
        {
            //Console.WriteLine("Move Command: " + _moveShapeCounter);
            //Console.WriteLine("Throttled Move Command: " + _throttledMoveShapeCounter);
            //Console.WriteLine("Delay Storyboard Completed: " + _delayStoryboardCompletedCounter);
            //_moveShapeCounter = 0;
            //_delayStoryboardCompletedCounter = 0;
            //_throttledMoveShapeCounter = 0;
            ////(Model as ZPortalShape).InvalidateShape();
        }


        #endregion

        #region Move Shape Command

        private RoutedCommand _moveShapeCommand;
        public RoutedCommand MoveShapeCommand
        {
            get
            {
                return _moveShapeCommand;
            }
            set
            {
                _moveShapeCommand = value;
                RaisePropertyChangedEvent("MoveShapeCommand");
            }
        }

        private void CanExecuteMoveShapeCommand(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = true;
        }

        private void ExecuteMoveShapeCommand(object sender, ExecutedRoutedEventArgs e)
        {
            _shapeMoves.OnNext(new Unit());
        }

        #endregion

        #region ElementAddedCommand
        private RoutedCommand _elementAddedCommand;
        public RoutedCommand ElementAddedCommand
        {
            get
            {
                return _elementAddedCommand;
            }
            set
            {
                _elementAddedCommand = value;
                RaisePropertyChangedEvent("ElementAddedCommand");
            }
        }

        private void ExecuteElementAddedCommand(object sender, ExecutedRoutedEventArgs e)
        {
            InitShape();
            RaisePropertyChangedEvent("SelectedElements");
        }

        private void CanExecuteElementAddedCommand(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = true;
        }

        #endregion

        #endregion
    }
}
