﻿using System;
using System.ComponentModel;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using Db4objects.Db4o.Linq;
using ZOIL.DataBackend;
using ZOIL.DataModel;
using ZOIL.Library.Behaviors;

namespace ZOIL.Library.Components
{
    [ViewModelFor(typeof(ZPortalView))]
    public class ZPortalViewModel : BaseViewModel, IConnectionShape
    {
        #region ctor

        public ZPortalViewModel()
        {
            initCommandBindings();
            registerChangeNotifications();
            ConnectionShape = new RectangleGeometry();
        }

        #endregion

        #region private fields

        private IDisposable _shapePropertyChangedSubscription;

        #endregion

        #region private methods

        private void registerChangeNotifications()
        {
            RegisterChangeNotification("IsMagicLense", "IsMagicLense");
        }

        private void initCommandBindings()
        {
            this.ClosePortalCommand = new RoutedCommand("ClosePortalCommand", typeof(ZPortalViewModel));
            CommandBinding cb2 = new CommandBinding(ClosePortalCommand, ExecuteClosePortalCommand, CanExecuteClosePortalCommand);

            this.MovePortalCommand = new RoutedCommand("MovePortalCommand", typeof(ZPortalViewModel));
            CommandBinding cb = new CommandBinding(MovePortalCommand, ExecuteMovePortalCommand, CanExecuteMovePortalCommand);

            this.ElementAddedCommand = new RoutedCommand("ElementAddedCommand", typeof(ZPortalViewModel));
            CommandBinding cb3 = new CommandBinding(ElementAddedCommand, ExecuteElementAddedCommand, CanExecuteElementAddedCommand);

            this.ElementRemovedCommand = new RoutedCommand("ElementRemovedCommand", typeof(ZPortalViewModel));
            CommandBinding cb4 = new CommandBinding(ElementRemovedCommand, ExecuteElementRemovedCommand, CanExecuteElementRemovedCommand);


            this.CommandBindings.Add(cb2);
            this.CommandBindings.Add(cb);
            this.CommandBindings.Add(cb3);
            this.CommandBindings.Add(cb4);
        }

        private void UpdateConnectionShape()
        {
            ConnectionShape = new RectangleGeometry(new Rect(new Point(Canvas.GetLeft(View), Canvas.GetTop(View)), new Size(View.Width, View.Height)));
        }

        #endregion

        #region public properties

        #region PortalShape

        public ZPortalShape PortalShape
        {
            get
            {
                return (Model as ZPortal).PortalShape;
            }
            set
            {
                (Model as ZPortal).PortalShape = value;
                RaisePropertyChangedEvent("PortalShape");
            }
        }

        #endregion

        #region IsMagicLense

        public bool IsMagicLense
        {
            get
            {
                return (Model as ZPortal).IsMagicLense;
            }
            set
            {
                (Model as ZPortal).IsMagicLense = value;
                RaisePropertyChangedEvent("IsMagicLense");
            }
        }

        #endregion

        #region SelectedElements

        public DObservableCollection<Object> SelectedElements
        {
            get
            {
                return PortalShape.SelectedElements;
            }
        }

        #endregion

        #region ConnectionShape

        private Geometry _connectionShape;

        /// <summary>
        /// Gets or sets the ConnectionShape property. This observable property 
        /// indicates
        /// </summary>
        public Geometry ConnectionShape
        {
            get { return _connectionShape; }
            set
            {
                if (_connectionShape != value)
                {
                    Geometry old = _connectionShape;
                    _connectionShape = value;
                    OnConnectionShapeChanged(old, value);
                    RaisePropertyChangedEvent("ConnectionShape");
                }
            }
        }

        /// <summary>
        /// Provides derived classes an opportunity to handle changes to the ConnectionShape property.
        /// </summary>
        protected virtual void OnConnectionShapeChanged(Geometry oldConnectionShape, Geometry newConnectionShape)
        {
        }

        #endregion

        #endregion

        #region commands

        #region Close Portal Command

        private RoutedCommand _closePortalCommand;
        public RoutedCommand ClosePortalCommand
        {
            get
            {
                return _closePortalCommand;
            }
            set
            {
                _closePortalCommand = value;
                RaisePropertyChangedEvent("ClosePortalCommand");
            }
        }

        private void ExecuteClosePortalCommand(object sender, ExecutedRoutedEventArgs e)
        {
            var rootCollections = from RootCollection root in Database.getInstance().Objects
                                  select root;

            RootCollection rootCollection = rootCollections.Single();

            rootCollection.Remove((Model as ZPortal).PortalShape);
            
            rootCollection.Remove(Model as ZPortal);

            if (_shapePropertyChangedSubscription != null)
                _shapePropertyChangedSubscription.Dispose();
        }

        private void CanExecuteClosePortalCommand(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = true;
        }

        #endregion

        #region ElementAddedCommand

        private RoutedCommand _elementAddedCommand;
        public RoutedCommand ElementAddedCommand
        {
            get
            {
                return _elementAddedCommand;
            }
            set
            {
                _elementAddedCommand = value;
                RaisePropertyChangedEvent("ElementAddedCommand");
            }
        }

        private void ExecuteElementAddedCommand(object sender, ExecutedRoutedEventArgs e)
        {
            OnPortalAdded();
        }

        private void CanExecuteElementAddedCommand(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = true;
        }

        #endregion

        #region ElementRemovedCommand

        private RoutedCommand _elementRemovedCommand;
        public RoutedCommand ElementRemovedCommand
        {
            get
            {
                return _elementRemovedCommand;
            }
            set
            {
                _elementRemovedCommand = value;
                RaisePropertyChangedEvent("ElementRemovedCommand");
            }
        }

        private void ExecuteElementRemovedCommand(object sender, ExecutedRoutedEventArgs e)
        {
            
        }

        private void CanExecuteElementRemovedCommand(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = true;
        }

        #endregion

        #region Move Portal Command

        private RoutedCommand _movePortalCommand;
        public RoutedCommand MovePortalCommand
        {
            get
            {
                return _movePortalCommand;
            }
            set
            {
                _movePortalCommand = value;
                RaisePropertyChangedEvent("MovePortalCommand");
            }
        }

        private void CanExecuteMovePortalCommand(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = true;
        }

        private void ExecuteMovePortalCommand(object sender, ExecutedRoutedEventArgs e)
        {
            UpdateConnectionShape();
        }

        #endregion

        #endregion

        #region event handling

        private void OnPortalAdded()
        {
            var rootCollections = from RootCollection rc in Database.getInstance().Objects
                                  select rc;

            RootCollection _rootCollection = rootCollections.Single();

            ZPortal portal = Model as ZPortal;
            ZPortalShape shape = portal.PortalShape;

            var propertyChanged = Observable.FromEvent<PropertyChangedEventArgs>(shape, "PropertyChanged").Where(evt => evt.EventArgs.PropertyName == "SelectedElements");
            _shapePropertyChangedSubscription = propertyChanged.Subscribe((evt) => RaisePropertyChangedEvent("SelectedElements"));

            if (IsMagicLense)
            {
                IVisualProperties shapeProp = _rootCollection.GetVisualProperties(shape.ID).First();
                shapeProp.Width = shape.Bounds.Width;
                shapeProp.Height = shape.Bounds.Height;
                shapeProp.X = shape.Bounds.Left;
                shapeProp.Y = shape.Bounds.Top;
                shapeProp.TemplateId = "MagicLensePortalShapeTemplate";

                IVisualProperties portalProp = _rootCollection.GetVisualProperties(portal.ID).First();
                portalProp.Width = shape.Bounds.Width;
                portalProp.Height = shape.Bounds.Height;
                portalProp.X = shape.Bounds.Left;
                portalProp.Y = shape.Bounds.Top;
            }
            else
            {
                if (!portal.IsInitialized)
                {
                    IVisualProperties portalProp = _rootCollection.GetVisualProperties(portal.ID).First();
                    
                    //place portal 100px next to shape
                    portalProp.X = shape.Bounds.Left + shape.Bounds.Width + 100;
                    portalProp.Y = shape.Bounds.Top;
                    portal.IsInitialized = true;
                }
            }
            RaisePropertyChangedEvent("SelectedElements");
            UpdateConnectionShape();
        }

        #endregion
    }
}
