﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Windows;
using System.Windows.Input;
using ZOIL.Framework.Core;
using ZOIL.Framework.Presentation;

namespace ZOIL.Library.Devices
{
    /// <summary>
    /// Default Handler for the Mouse Device
    /// </summary>
    public class ZLandscapeDefaultMouseHandler : ZLandscapeInputHandler
    {
        #region private fields

        private IEnumerable<ZInformationLandscape> _parentLandscapes;
        
        private bool _isPanning = false;
        
        private bool _mouseDown = false;
        
        private Point _lastMousePosition;
        
        private Point _lastMouseDownPosition;
        
        private bool _animateWheelZoom;
        
        private double _wheelZoomFactor;

        #endregion

        #region public properties

        #region AnimateWheelZoom

        /// <summary>
        /// specifies if zooming with the mouse wheel is animated or not
        /// </summary>
        public bool AnimateWheelZoom
        {
            get
            {
                return this._animateWheelZoom;
            }
            set
            {
                this._animateWheelZoom = value;
                OnPropertyChanged("AnimateWheelZoom");
            }
        }

        #endregion

        #region WheelZoomFactor

        /// <summary>
        /// specifies the factor by which the landscape is zoomed using the mouse wheel
        /// </summary>
        public double WheelZoomFactor
        {
            get
            {
                return this._wheelZoomFactor;
            }
            set
            {
                this._wheelZoomFactor = value;
                OnPropertyChanged("WheelZoomFactor");
            }
        }

        #endregion

        #endregion

        #region ctor

        /// <summary>
        /// Constructor of the default handler for the mouse device
        /// </summary>
        /// <param name="landscape"></param>
        public ZLandscapeDefaultMouseHandler(ZInformationLandscape landscape)
            : base(landscape)
        {
            AnimateWheelZoom = false;
            WheelZoomFactor = 0.95;
            IsEnabled = true;
        }

        #endregion

        #region overrides

        /// <summary>
        /// Called whenever the value of <see cref="ZLandscapeInputHandler.IsPanningEnabled"/> has changed
        /// </summary>
        protected override void OnIsPanningEnabledChanged()
        {
            if (IsPanningEnabled)
            {
                Mouse.AddMouseUpHandler(this.InformationLandscape, OnMouseButton);
                Mouse.AddMouseDownHandler(this.InformationLandscape, OnMouseButton);
                Mouse.AddMouseMoveHandler(this.InformationLandscape, OnMouseMove);
            }
            else
            {
                Mouse.RemoveMouseUpHandler(this.InformationLandscape, OnMouseButton);
                Mouse.RemoveMouseDownHandler(this.InformationLandscape, OnMouseButton);
                Mouse.RemoveMouseMoveHandler(this.InformationLandscape, OnMouseMove);
            }
        }

        /// <summary>
        /// Called whenever the value of <see cref="ZLandscapeInputHandler.IsZoomingEnabled"/> has changed
        /// </summary>
        protected override void OnIsZoomingEnabledChanged()
        {
            if (IsZoomingEnabled)
            {
                Mouse.AddMouseWheelHandler(this.InformationLandscape, OnMouseWheel);
            }
            else
            {
                Mouse.RemoveMouseWheelHandler(this.InformationLandscape, OnMouseWheel);
            }
        }

        #endregion

        #region event handling

        private void OnMouseWheel(object sender, MouseWheelEventArgs e)
        {
            if (e.Delta < 0)
            {
                this.InformationLandscape.ZoomBy(Mouse.GetPosition(this.InformationLandscape), WheelZoomFactor, AnimateWheelZoom);
            }
            else
            {
                this.InformationLandscape.ZoomBy(Mouse.GetPosition(this.InformationLandscape), 1.0 / WheelZoomFactor, AnimateWheelZoom);
            }
        }

        private void OnMouseButton(object sender, MouseButtonEventArgs e)
        {
            // ignore input of stylus devices, since we have another handler for this
            if (e.StylusDevice != null)
                return;

#if DEBUG
            ZDebug.QuickLog("HandleMouseButton received: " + e.RoutedEvent.ToString());
#endif
            if (e.ButtonState == MouseButtonState.Pressed)
            {
                _mouseDown = true;

                _parentLandscapes = ZTreeHelper.TryFindParents<ZInformationLandscape>(InformationLandscape);
                _lastMouseDownPosition = e.GetPosition(null);
                InformationLandscape.CaptureMouse();

                e.Handled = true;

            }
            else if (e.ButtonState == MouseButtonState.Released)
            {
                InformationLandscape.ReleaseMouseCapture();
                if (_isPanning)
                {
                    _isPanning = false;
                    e.Handled = true;
                    _mouseDown = false;
                    return;
                }

                FrameworkElement clickedElement = Mouse.DirectlyOver as FrameworkElement;
                if (clickedElement != null)
                {
#if DEBUG
                    bool elementIsZoomTarget = ZInformationLandscape.GetZoomTarget(clickedElement);
                    ZDebug.QuickLog("HandleMouseButton MouseUp over " + clickedElement + " elementIsZoomTarget=" + elementIsZoomTarget);
#endif
                    InformationLandscape.ZoomOnto(clickedElement, false, true);
                    clickedElement.RaiseEvent(new RoutedEventArgs(ZInformationLandscape.ClickZoomEvent, clickedElement));
                    _mouseDown = false;
                    e.Handled = true;
                }

            }
        }

        private void OnMouseMove(object sender, MouseEventArgs e)
        {
            if (_isPanning && _mouseDown)
            {
                Point curPoint = Mouse.GetPosition(null);
                double scale = this.InformationLandscape.Animation.Position.Zoom;

                Point landscapePoint = Mouse.GetPosition(InformationLandscape);

                // look for all parent Informationlandscapes and adjust the scale factor
                foreach (ZInformationLandscape cur in _parentLandscapes)
                {
                    double scale2 = cur.Animation.Position.Zoom;
                    scale *= scale2;
                }

                double dx = (curPoint.X - _lastMousePosition.X) / scale;
                double dy = (curPoint.Y - _lastMousePosition.Y) / scale;
                Vector v = new Vector(dx, dy);

                InformationLandscape.PanBy(v, false);
            }
            else if (_mouseDown && (Math.Abs(_lastMouseDownPosition.X - e.GetPosition(null).X) > 10 || Math.Abs(_lastMouseDownPosition.Y - e.GetPosition(null).Y) > 10))
            {
                _isPanning = true;
            }

            _lastMousePosition = e.GetPosition(null);
        }

        #endregion

        
    }
}
