﻿using System;
using System.Collections.Generic;
using System.Windows;
using Microsoft.Surface.Presentation;
using Microsoft.Surface.Presentation.Manipulations;
using ZOIL.Framework.Core;
using ZOIL.Framework.Presentation;


namespace ZOIL.Library.Devices.Surface
{
    /// <summary>
    /// Default Handler for the Microsoft Surface Device
    /// </summary>
    public class ZLandscapeDefaultSurfaceHandler : ZLandscapeInputHandler
    {
        #region private fields

        private Affine2DManipulationProcessor _scaleProcessor;

        private Affine2DManipulationProcessor _panProcessor;

        private IEnumerable<ZInformationLandscape> _parentLandscapes;

        private Point _scaleManipulationOrigin;

        #endregion

        #region public properties

        /// <summary>
        /// Gets the origin of the current scale manipulation
        /// </summary>
        public Point ScaleManipulationOrigin
        {
            get
            {
                return _scaleManipulationOrigin;
            }
            private set
            {
                _scaleManipulationOrigin = value;
                OnPropertyChanged("ScaleManipulationOrigin");
            }
        }

        #endregion

        #region ctor

        /// <summary>
        /// Constructor of the default handler for the Microsoft Surface Device
        /// </summary>
        /// <param name="landscape"></param>
        public ZLandscapeDefaultSurfaceHandler(ZInformationLandscape landscape)
            : base(landscape)
        {
            //Contacts.EnableInput(CurrentWindow);

            Contacts.AddContactDownHandler(InformationLandscape, OnContactDown);
            Contacts.AddContactUpHandler(InformationLandscape, OnContactUp);

            _panProcessor = new Affine2DManipulationProcessor(Affine2DManipulations.TranslateX | Affine2DManipulations.TranslateY, InformationLandscape, true);
            _scaleProcessor = new Affine2DManipulationProcessor(Affine2DManipulations.Scale, InformationLandscape, true);
            IsEnabled = true;
        }

        #endregion

        #region Panning

        void OnPanDelta(object sender, Affine2DOperationDeltaEventArgs e)
        {
            double scale = this.InformationLandscape.Animation.Position.Zoom;

            // look for all parent Informationlandscapes and adjust the scale factor
            foreach (ZInformationLandscape cur in _parentLandscapes)
            {
                double scale2 = cur.Animation.Position.Zoom;
                scale *= scale2;
            }

            InformationLandscape.PanBy(new Vector(e.Delta.X / scale, e.Delta.Y / scale), false);

#if DEBUG
            ZDebug.QuickLog("Panning with global scale factor: " + scale);
#endif
        }

        #endregion

        #region Scaling

        private void OnScaleDelta(object sender, Affine2DOperationDeltaEventArgs e)
        {
#if DEBUG
            ZDebug.QuickLog("Zooming with ManipulationOrigin: " + e.ManipulationOrigin);
#endif
            ScaleManipulationOrigin = e.ManipulationOrigin;
            InformationLandscape.ZoomBy(CurrentWindow.TranslatePoint(e.ManipulationOrigin, InformationLandscape), e.ScaleDelta, false);
        }

        #endregion

        #region Click Zoom

        private FrameworkElement _clickZoomSender;
        private FrameworkElement _clickZoomSource;
        Dictionary<IInputElement, Point> _clickZoomRoots = new Dictionary<IInputElement, Point>();

        private void DoClickZoom()
        {
            FrameworkElement zoomTarget = ZTreeHelper.GetClosestZoomTarget(_clickZoomSender, _clickZoomSource);
            if (zoomTarget != null)
            {
                InformationLandscape.ZoomOnto(zoomTarget, false, true);
                zoomTarget.RaiseEvent(new RoutedEventArgs(ZInformationLandscape.ClickZoomEvent, zoomTarget));
            }
        }

        #endregion

        #region overrides

        /// <summary>
        /// Registers or deregisters the event handlers which are responsible for panning
        /// </summary>
        protected override void OnIsPanningEnabledChanged()
        {
            if (IsPanningEnabled)
            {
                _panProcessor.Affine2DManipulationDelta -= OnPanDelta;
                _panProcessor.Affine2DManipulationDelta += OnPanDelta;
            }
            else
            {
                _panProcessor.Affine2DManipulationDelta -= OnPanDelta;
            }
        }

        /// <summary>
        /// Registers or deregisters the event handlers which are responsible for zooming
        /// </summary>
        protected override void OnIsZoomingEnabledChanged()
        {
            if (IsZoomingEnabled)
            {
                _scaleProcessor.Affine2DManipulationDelta -= OnScaleDelta;
                _scaleProcessor.Affine2DManipulationDelta += OnScaleDelta;
            }
            else
            {
                _scaleProcessor.Affine2DManipulationDelta -= OnScaleDelta;
            }
        }

        #endregion

        #region Contact Handler

        private void OnContactDown(Object sender, ContactEventArgs e)
        {
            if (!e.Contact.IsFingerRecognized)
                return;

            Point position = e.GetPosition(null);

            if (IsClickZoomEnabled)
            {
                _clickZoomSender = sender as FrameworkElement;
                _clickZoomSource = e.Source as FrameworkElement;
                if (_clickZoomRoots.ContainsKey(e.Contact.DirectlyOver))
                    _clickZoomRoots.Remove(e.Contact.DirectlyOver);
                _clickZoomRoots.Add(e.Contact.DirectlyOver, e.GetPosition(null));
            }

            _parentLandscapes = ZTreeHelper.TryFindParents<ZInformationLandscape>(InformationLandscape);

            _scaleProcessor.BeginTrack(e.Contact);

            _panProcessor.BeginTrack(e.Contact);

            if (IsEnabled)
            {
                e.Handled = true;
            }
        }

        private void OnContactUp(Object sender, ContactEventArgs e)
        {
            if (!e.Contact.IsFingerRecognized)
                return;

            _scaleProcessor.EndTrack(e.Contact);
            _panProcessor.EndTrack(e.Contact);

            // if we didn't pan over a certain amount of pixels, let's do a click zoom
            if (IsClickZoomEnabled)
            {
                if (_clickZoomRoots.ContainsKey(e.Contact.DirectlyOver))
                {
                    double panDistance = Point.Subtract(e.GetPosition(null), _clickZoomRoots[e.Contact.DirectlyOver]).Length;
                    if (panDistance < 10)
                    {
                        DoClickZoom();
                    }
                    _clickZoomRoots.Remove(e.Contact.DirectlyOver);
                }
            }

            if (IsEnabled)
            {
                e.Handled = true;
            }
        }

        #endregion

    }
}
