﻿using System;
using System.ComponentModel;
using System.Windows;
using ZOIL.Framework.Presentation;

namespace ZOIL.Library.Devices
{
    /// <summary>
    /// An abstract base class for all input handlers that modify a <see cref="ZInformationLandscape"/>
    /// </summary>
    public abstract class ZLandscapeInputHandler : INotifyPropertyChanged
    {
        #region private fields

        private ZInformationLandscape _informationLandscape;
        
        private bool _isEnabled;
        
        private bool _isPanningEnabled;

        private bool _isZoomingEnabled;

        private bool _isClickZoomEnabled;
        
        #endregion

        #region public properties

        #region InformationLandscape

        /// <summary>
        /// Specifies the <see cref="ZInformationLandscape"/> that will be manipulated by this handler
        /// </summary>
        public ZInformationLandscape InformationLandscape
        {
            get
            {
                return _informationLandscape;
            }
            set
            {
                _informationLandscape = value;
                OnPropertyChanged("InformationLandscape");
            }
        }

        #endregion

        #region IsEnabled

        /// <summary>
        /// specifies if the handler is currently enabled to manipulate the landscape
        /// </summary>
        public bool IsEnabled
        {
            get
            {
                return _isEnabled;
            }
            set
            {
                _isEnabled = value;
                OnPropertyChanged("IsEnabled");
            }
        }

        #endregion

        #region IsPanningEnabled

        /// <summary>
        /// specifies if panning with this handler is currently enabled
        /// </summary>
        public bool IsPanningEnabled
        {
            get
            {
                return _isPanningEnabled;
            }
            set
            {
                _isPanningEnabled = value;
                OnPropertyChanged("IsPanningEnabled");
            }
        }

        #endregion

        #region IsZoomingEnabled

        /// <summary>
        /// specifies if zooming with this handler is currently enabled
        /// </summary>
        public bool IsZoomingEnabled
        {
            get
            {
                return _isZoomingEnabled;
            }
            set
            {
                _isZoomingEnabled = value;
                OnPropertyChanged("IsZoomingEnabled");
            }
        }

        #endregion

        #region IsClickZoomEnabled

        /// <summary>
        /// specifies if click zooms with this handler are currently enabled
        /// </summary>
        public bool IsClickZoomEnabled
        {
            get
            {
                return _isClickZoomEnabled;
            }
            set
            {
                _isClickZoomEnabled = value;
                OnPropertyChanged("IsClickZoomEnabled");
            }
        }

        #endregion

        #region CurrentWindow

        /// <summary>
        /// Gets the current Window in which the landscape lies
        /// </summary>
        protected Window CurrentWindow
        {
            get
            {
                return Window.GetWindow(InformationLandscape);
            }
        }

        #endregion

        #endregion

        #region ctor

        /// <summary>
        /// constructor of the abstract base class for all input handlers
        /// </summary>
        /// <param name="landscape"></param>
        protected ZLandscapeInputHandler(ZInformationLandscape landscape)
        {
            this.InformationLandscape = landscape;
            this.PropertyChanged += new PropertyChangedEventHandler(OnPropertyChanged);
        }

        #endregion

        #region property changed handler

        /// <summary>
        /// called whenever a property of the handler changes
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected virtual void OnPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e == null)
                throw new ArgumentNullException("e");


            switch (e.PropertyName)
            {
                case "IsEnabled":
                    OnIsEnabledChanged();
                    break;
                case "IsPanningEnabled":
                    OnIsPanningEnabledChanged();
                    break;
                case "IsZoomingEnabled":
                    OnIsZoomingEnabledChanged();
                    break;
                case "IsClickZoomEnabled":
                    OnIsClickZoomEnabledChanged();
                    break;
                default:
                    break;
            }
        }

        /// <summary>
        /// Called whenever the value of <see cref="IsEnabled"/> has changed. Switches on or off the properties
        /// <see cref="IsPanningEnabled"/>, <see cref="IsZoomingEnabled"/> and <see cref="IsClickZoomEnabled"/>
        /// </summary>
        protected virtual void OnIsEnabledChanged()
        {
            if (IsEnabled)
            {
                IsPanningEnabled = true;
                IsZoomingEnabled = true;
                IsClickZoomEnabled = true;
            }
            else
            {
                IsPanningEnabled = false;
                IsZoomingEnabled = false;
                IsClickZoomEnabled = false;
            }
        }

        /// <summary>
        /// Called whenever the value of <see cref="IsPanningEnabled"/> has changed
        /// </summary>
        protected virtual void OnIsPanningEnabledChanged()
        {

        }

        /// <summary>
        /// called whenever the value of <see cref="IsZoomingEnabled"/> has changed
        /// </summary>
        protected virtual void OnIsZoomingEnabledChanged()
        {

        }

        /// <summary>
        /// called whenever the value of <see cref="IsClickZoomEnabled"/> has changed
        /// </summary>
        protected virtual void OnIsClickZoomEnabledChanged()
        {

        }

        #endregion

        #region INotifyPropertyChanged Members

        /// <summary>
        /// default INotifyPropertyChanged implementation
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// raises the PropertyChanged event with the given property name
        /// </summary>
        /// <param name="propertyName"></param>
        protected void OnPropertyChanged(string propertyName)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
        }

        #endregion
    }
}
